from datetime import date, time
from json import loads
from typing import List

from dateutil.parser import isoparse
from typing_extensions import TypedDict

from elimity_insights_client._util import map_list
from elimity_insights_client.api.query_results_page import (
    QueryResultsPage,
    QueryResult,
    Entity,
    Value,
    BooleanValue,
    DateValue,
    DateTimeValue,
    NumberValue,
    StringValue,
    TimeValue,
    GroupByQueryResultsPage,
    GroupByQueryResult,
)


class EntityDict(TypedDict):
    active: bool
    id: str
    name: str


class GroupByQueryResultsPageDict(TypedDict):
    groupCount: int
    results: List["GroupByQueryResultDict"]


class GroupByQueryResultDict(TypedDict):
    count: int
    label: "ValueDict"
    subPages: List[GroupByQueryResultsPageDict]


class QueryResultsPageDict(TypedDict):
    count: int
    results: List["QueryResultDict"]


class QueryResultDict(TypedDict):
    entity: EntityDict
    inclusions: List["ValueDict"]
    linkGroupByPages: List[GroupByQueryResultsPageDict]
    linkPages: List[QueryResultsPageDict]


class ValueDict(TypedDict):
    type: str
    value: str


def decode_query_results_page(dict: QueryResultsPageDict) -> QueryResultsPage:
    count = dict["count"]
    result_dicts = dict["results"]
    results = map_list(_decode_query_result, result_dicts)
    return QueryResultsPage(count, results)


def _decode_group_by_query_result(dict: GroupByQueryResultDict) -> GroupByQueryResult:
    count = dict["count"]
    label_dict = dict["label"]
    label = _decode_value(label_dict)
    sub_page_dicts = dict["subPages"]
    sub_pages = map_list(_decode_group_by_query_results_page, sub_page_dicts)
    return GroupByQueryResult(count, label, sub_pages)


def _decode_group_by_query_results_page(
    dict: GroupByQueryResultsPageDict,
) -> GroupByQueryResultsPage:
    group_count = dict["groupCount"]
    result_dicts = dict["results"]
    results = map_list(_decode_group_by_query_result, result_dicts)
    return GroupByQueryResultsPage(group_count, results)


def _decode_query_result(dict: QueryResultDict) -> QueryResult:
    entity_dict = dict["entity"]
    active = entity_dict["active"]
    id = entity_dict["id"]
    name = entity_dict["name"]
    entity = Entity(active, id, name)
    inclusion_dicts = dict["inclusions"]
    inclusions = map_list(_decode_value, inclusion_dicts)
    link_group_by_page_dicts = dict["linkGroupByPages"]
    link_group_by_pages = map_list(
        _decode_group_by_query_results_page, link_group_by_page_dicts
    )
    link_page_dicts = dict["linkPages"]
    link_pages = map_list(decode_query_results_page, link_page_dicts)
    return QueryResult(entity, inclusions, link_group_by_pages, link_pages)


def _decode_value(dict: ValueDict) -> Value:
    type = dict["type"]
    value = dict["value"]
    if type == "boolean":
        return BooleanValue(value == "true")

    if type == "date":
        date_value = date.fromisoformat(value)
        return DateValue(date_value)

    if type == "dateTime":
        date_time_value = isoparse(value)
        return DateTimeValue(date_time_value)

    if type == "number":
        number_value = loads(value)
        return NumberValue(number_value)

    if type == "string":
        return StringValue(value)

    time_value = time.fromisoformat(value)
    return TimeValue(time_value)



from dataclasses import dataclass
from datetime import datetime
from enum import Enum, auto
from itertools import chain
from typing import Optional, Union, Iterable, Tuple, Dict
from zlib import compressobj

from requests import request, Response

from elimity_insights_client._decode_domain_graph_schema import (
    decode_domain_graph_schema,
)
from elimity_insights_client._domain_graph_schema import (
    DomainGraphSchema,
)
from elimity_insights_client._util import encode_datetime, encoder


@dataclass
class AttributeAssignment:
    """Assignment of a value for an attribute type."""

    attribute_type_id: str
    value: "Value"


@dataclass
class BooleanValue:
    """Value to assign for a boolean attribute type."""

    value: bool


@dataclass
class Certificate:
    """Client side certificate for mTLS connections."""

    certificate_path: str
    private_key_path: str


class Client:
    """Client for connector interactions with an Elimity Insights server."""

    def __init__(self, config: "Config") -> None:
        """Return a new client with the given configuration."""
        self._config = config

    def create_connector_logs(self, logs: Iterable["ConnectorLog"]) -> None:
        """Create connector logs."""
        json = map(_encode_connector_log, logs)
        json_string = encoder.encode(json)
        json_bytes = json_string.encode()
        self._post("application/json", json_bytes, "connector-logs")

    def get_domain_graph_schema(self) -> "DomainGraphSchema":
        """Retrieve the domain graph schema."""
        headers: Dict[str, str] = {}
        response = self._request(None, headers, "GET", "domain-graph-schema")
        json = response.json()
        return decode_domain_graph_schema(json)

    def reload_domain_graph(self, graph: "DomainGraph") -> None:
        """
        Reload a domain graph.

        This method serializes the given domain graph by streaming its entities
        and relationships to a compressed buffer. It always exhausts the given
        domain graph's entities before iterating its relationships.
        """
        json = _encode_domain_graph(graph)
        json_bytes_chunks = _compress_domain_graph(json)
        json_bytes_iter = chain.from_iterable(json_bytes_chunks)
        json_bytes = bytes(json_bytes_iter)
        self._post("application/octet-stream", json_bytes, "snapshots")

    def _post(self, content_type: str, data: bytes, path: str) -> None:
        headers = {"Content-Type": content_type}
        self._request(data, headers, "POST", path)

    def _request(
        self,
        data: Optional[bytes],
        headers: Dict[str, str],
        method: str,
        path: str,
    ) -> Response:
        config = self._config
        id_ = config.id
        url = f"{config.url}/api/custom-sources/{id_}/{path}"
        auth = str(id_), config.token
        cert = _cert(config.certificate)
        response = request(
            method,
            url,
            auth=auth,
            cert=cert,
            data=data,
            headers=headers,
            verify=config.verify_ssl,
        )
        response.raise_for_status()
        return response


@dataclass
class Config:
    """Configuration for an Elimity Insights client."""

    id: int
    url: str
    token: str
    verify_ssl: bool = True
    certificate: Optional[Certificate] = None


@dataclass
class ConnectorLog:
    """Log line produced by an Elimity Insights connector."""

    level: "Level"
    message: str
    timestamp: datetime


@dataclass
class DateTime:
    """Date-time in UTC."""

    year: int
    month: int
    day: int
    hour: int
    minute: int
    second: int


@dataclass
class DateTimeValue:
    """Value to assign for a date-time attribute type, in UTC."""

    value: DateTime


@dataclass
class DateValue:
    """Value to assign for a date attribute type."""

    year: int
    month: int
    day: int


@dataclass
class DomainGraph:
    """Snapshot of a complete domain graph at a specific timestamp."""

    entities: Iterable["Entity"]
    relationships: Iterable["Relationship"]
    timestamp: Optional[DateTime] = None


@dataclass
class Entity:
    """Entity of a specific type, including attribute assignments."""

    attribute_assignments: Iterable[AttributeAssignment]
    id: str
    name: str
    type: str


class Level(Enum):
    """Severity level of an Elimity Insights connector log line."""

    ALERT = auto()
    INFO = auto()


@dataclass
class NumberValue:
    """Value to assign for a number attribute type."""

    value: float


@dataclass
class Relationship:
    """Relationship between two entities, including attribute assignments."""

    attribute_assignments: Iterable[AttributeAssignment]
    from_entity_id: str
    from_entity_type: str
    to_entity_id: str
    to_entity_type: str


@dataclass
class StringValue:
    """Value to assign for a string attribute type."""

    value: str


@dataclass
class TimeValue:
    """Value to assign for a time attribute type, in UTC."""

    hour: int
    minute: int
    second: int


Value = Union[
    BooleanValue, DateValue, DateTimeValue, NumberValue, StringValue, TimeValue
]


def _cert(certificate: Optional[Certificate]) -> Optional[Tuple[str, str]]:
    if certificate is None:
        return None
    else:
        return certificate.certificate_path, certificate.private_key_path


def _compress_domain_graph(json: object) -> Iterable[bytes]:
    compress = compressobj()
    for json_string_chunk in encoder.iterencode(json, False):
        json_bytes_chunk = json_string_chunk.encode()
        yield compress.compress(json_bytes_chunk)
    yield compress.flush()


def _encode_attribute_assignment(assignment: AttributeAssignment) -> object:
    value = _encode_value(assignment.value)
    return {
        "attributeTypeId": assignment.attribute_type_id,
        "value": value,
    }


def _encode_connector_log(log: ConnectorLog) -> object:
    level = _encode_level(log.level)
    timestamp = encode_datetime(log.timestamp)
    return {
        "level": level,
        "message": log.message,
        "timestamp": timestamp,
    }


def _encode_date(year: int, month: int, day: int) -> object:
    return {"year": year, "month": month, "day": day}


def _encode_date_time(time: DateTime) -> object:
    return {
        "year": time.year,
        "month": time.month,
        "day": time.day,
        "hour": time.hour,
        "minute": time.minute,
        "second": time.second,
    }


def _encode_domain_graph(graph: DomainGraph) -> object:
    entities = map(_encode_entity, graph.entities)
    relationships = map(_encode_relationship, graph.relationships)
    obj = {"entities": entities, "relationships": relationships}
    if graph.timestamp is None:
        return obj
    else:
        history_timestamp = _encode_date_time(graph.timestamp)
        return {**obj, "historyTimestamp": history_timestamp}


def _encode_entity(entity: Entity) -> object:
    assignments = map(_encode_attribute_assignment, entity.attribute_assignments)
    return {
        "attributeAssignments": assignments,
        "id": entity.id,
        "name": entity.name,
        "type": entity.type,
    }


def _encode_level(level: Level) -> object:
    if level == Level.ALERT:
        return "alert"
    else:
        return "info"


def _encode_relationship(relationship: Relationship) -> object:
    assignments = map(_encode_attribute_assignment, relationship.attribute_assignments)
    return {
        "attributeAssignments": assignments,
        "fromEntityId": relationship.from_entity_id,
        "toEntityId": relationship.to_entity_id,
        "fromEntityType": relationship.from_entity_type,
        "toEntityType": relationship.to_entity_type,
    }


def _encode_time(hour: int, minute: int, second: int) -> object:
    return {"hour": hour, "minute": minute, "second": second}


def _encode_value(value: Value) -> object:
    if isinstance(value, BooleanValue):
        return {"type": "boolean", "value": value.value}

    elif isinstance(value, DateValue):
        date_value = _encode_date(value.year, value.month, value.day)
        return {"type": "date", "value": date_value}

    elif isinstance(value, DateTimeValue):
        date_time_value = _encode_date_time(value.value)
        return {"type": "dateTime", "value": date_time_value}

    elif isinstance(value, NumberValue):
        return {"type": "number", "value": value.value}

    elif isinstance(value, StringValue):
        return {"type": "string", "value": value.value}

    else:
        time_value = _encode_time(value.hour, value.minute, value.second)
        return {"type": "time", "value": time_value}



"""Type definitions for queries."""

from dataclasses import dataclass
from enum import Enum, auto
from typing import Union, List

from elimity_insights_client.api.expression import (
    BooleanExpression,
    DateExpression,
    DateTimeExpression,
    NumberExpression,
    StringExpression,
    TimeExpression,
)


@dataclass
class BooleanAnyExpression:
    """Untyped wrapper for a boolean expression."""

    expr: BooleanExpression


@dataclass
class DateAnyExpression:
    """Untyped wrapper for a date expression."""

    expr: DateExpression


@dataclass
class DateTimeAnyExpression:
    """Untyped wrapper for a date-time expression."""

    expr: DateTimeExpression


@dataclass
class DirectedGroupOrdering:
    """Group ordering according to a given direction and type."""

    direction: "Direction"
    type: "GroupOrderingType"


class Direction(Enum):
    """Direction for ordering query results."""

    ASC = auto()
    DESC = auto()


@dataclass
class DirectLinkGroupByQuery:
    """Subquery targeting groups of directly linked entities."""

    alias: str
    condition: BooleanExpression
    entity_type: str
    group_by: List["Grouping"]
    link_alias: str
    source_id: int


@dataclass
class DirectLinkQuery:
    """Subquery targeting directly linked entities."""

    alias: str
    condition: BooleanExpression
    direct_link_group_by_queries: List[DirectLinkGroupByQuery]
    direct_link_queries: List["DirectLinkQuery"]
    entity_type: str
    include: List["AnyExpression"]
    limit: int
    link_alias: str
    link_group_by_queries: List["LinkGroupByQuery"]
    link_queries: List["Query"]
    offset: int
    order_by: List["Ordering"]
    source_id: int


@dataclass
class Grouping:
    """Query clause indicating how to group results."""

    key: "AnyExpression"
    ordering: "GroupOrdering"


class GroupOrderingType(Enum):
    """Type indicating which property to use for ordering result groups."""

    COUNT = auto()
    LABEL = auto()


@dataclass
class LinkGroupByQuery:
    """Subquery targeting groups of linked entities."""

    alias: str
    condition: BooleanExpression
    entity_type: str
    group_by: List[Grouping]
    source_id: int


@dataclass
class NumberAnyExpression:
    """Untyped wrapper for a number expression."""

    expr: NumberExpression


@dataclass
class Ordering:
    """Query clause indicating how to order results."""

    any_expression: "AnyExpression"
    direction: Direction


@dataclass
class Query:
    """Base query targeting linked or top-level entities."""

    alias: str
    condition: BooleanExpression
    direct_link_group_by_queries: List[DirectLinkGroupByQuery]
    direct_link_queries: List[DirectLinkQuery]
    entity_type: str
    include: List["AnyExpression"]
    limit: int
    link_group_by_queries: List[LinkGroupByQuery]
    link_queries: List["Query"]
    offset: int
    order_by: List[Ordering]
    source_id: int


@dataclass
class StringAnyExpression:
    """Untyped wrapper for a string expression."""

    expr: StringExpression


@dataclass
class TimeAnyExpression:
    """Untyped wrapper for a time expression."""

    expr: TimeExpression


AnyExpression = Union[
    BooleanAnyExpression,
    DateAnyExpression,
    DateTimeAnyExpression,
    NumberAnyExpression,
    StringAnyExpression,
    TimeAnyExpression,
]


@dataclass
class UnspecifiedGroupOrdering:
    """Sentinel value representing an unspecified group ordering."""

    pass


GroupOrdering = Union[DirectedGroupOrdering, UnspecifiedGroupOrdering]



from dataclasses import dataclass
from enum import Enum, auto
from typing import List


@dataclass
class AttributeType:
    """Attribute type for an entity type."""

    archived: bool
    description: str
    entity_type: str
    id: str
    name: str
    type: "Type"


@dataclass
class DomainGraphSchema:
    """Schema determining valid domain graphs."""

    attribute_types: List[AttributeType]
    entity_types: List["EntityType"]
    relationship_attribute_types: List["RelationshipAttributeType"]


@dataclass
class EntityType:
    """Type of an entity."""

    anonymized: bool
    icon: str
    id: str
    plural: str
    singular: str


@dataclass
class RelationshipAttributeType:
    """Attribute type for relationships between entities of specific types."""

    archived: bool
    description: str
    from_entity_type: str
    id: str
    name: str
    to_entity_type: str
    type: "Type"


class Type(Enum):
    """Type of an attribute type, determining valid assignment values."""

    BOOLEAN = auto()
    DATE = auto()
    DATE_TIME = auto()
    NUMBER = auto()
    STRING = auto()
    TIME = auto()



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['elimity_insights_client', 'elimity_insights_client.api']

package_data = \
{'': ['*']}

install_requires = \
['python-dateutil>=2.8.2,<3.0.0',
 'requests>=2.28.1,<3.0.0',
 'simplejson>=3.17.6,<4.0.0',
 'typing-extensions>=4.3.0,<5.0.0']

setup_kwargs = {
    'name': 'elimity-insights-client',
    'version': '9.0.0b0',
    'description': 'Client for connector interactions with an Elimity Insights server',
    'long_description': '# Elimity Insights Python client\n\nThis Python module provides a client for connector interactions with an Elimity\nInsights server.\n\n## Usage\n\n### Importing data to custom sources\n\nThe following snippet shows how to authenticate as a custom source and create a connector log at an Elimity Insights\nserver. You can generate a source identifier and token by visiting the custom source\'s detail page in Elimity Insights\nand clicking the \'GENERATE CREDENTIALS\' button, which can be found under the \'SETTINGS\' tab.\n\n```python3\nfrom datetime import datetime\n\nfrom elimity_insights_client import Client, Config, ConnectorLog, Level\n\nif __name__ == "__main__":\n    config = Config(id=1, url="https://local.elimity.com:8081", token="token")\n    client = Client(config)\n\n    timestamp = datetime.now()\n    log = ConnectorLog(level=Level.INFO, message="Hello world!", timestamp=timestamp)\n    logs = [log]\n    client.create_connector_logs(logs)\n```\n\n### Other API interactions\n\nThis module also provides a client for other API interactions with Elimity Insights. The snippet below shows how to\nauthenticate with an API token and list sources at an Elimity Insights server. You can generate a token identifier and\nsecret by visiting the \'API tokens\' page in Elimity Insights and clicking the \'CREATE API TOKEN\' button.\n\n```python3\nfrom elimity_insights_client.api import Config, sources\n\nif __name__ == "__main__":\n    config = Config(token_id="1", token_secret="my-secret-value", url="https://example.elimity.com", verify_ssl=True)\n    my_sources = sources(config)\n    print(my_sources)\n```\n\n## Installation\n\n```sh\n$ pip install elimity-insights-client\n```\n\n## Compatibility\n\n| Client version | Insights version |\n| -------------- | ---------------- |\n| 1              | 2.8 - 2.10       |\n| 2 - 3          | 2.11 - 3.0       |\n| 4              | 3.1 - 3.3        |\n| 5 - 6          | 3.4 - 3.5        |\n| 7              | 3.6 - 3.7        |\n| 8              | 3.8 - 3.15       |\n| 9              | ^3.16            |\n',
    'author': 'Elimity development team',
    'author_email': 'dev@elimity.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/elimity-com/insights-client-python',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""Type definitions for sources."""

from dataclasses import dataclass
from datetime import datetime
from typing import Union

from elimity_insights_client._domain_graph_schema import DomainGraphSchema


@dataclass
class AbsentLastReloadTimestamp:
    """Sentinel value indicating a last reload timestamp is not applicable."""

    pass


@dataclass
class PresentLastReloadTimestamp:
    """Timestamp indicating when a source has been most recently reloaded."""

    value: datetime


LastReloadTimestamp = Union[AbsentLastReloadTimestamp, PresentLastReloadTimestamp]


@dataclass
class Source:
    """Source providing domain data to an Elimity Insights server."""

    archived: bool
    domain_graph_schema: DomainGraphSchema
    id: int
    last_reload_timestamp: LastReloadTimestamp
    name: str



"""Type definitions for expressions."""

from dataclasses import dataclass
from datetime import date, datetime, time
from enum import Enum, auto
from typing import Union, List


@dataclass
class ActiveBooleanExpression:
    """Expression evaluating whether a given entity is active."""

    reference: str


class AggregateOperator(Enum):
    """Operator to use when aggregating numeric expression results across entities."""

    AVG = auto()
    COUNT = auto()
    MAX = auto()
    MIN = auto()
    SUM = auto()


@dataclass
class AllBooleanExpression:
    """Expression evaluating whether all given sub-expressions hold."""

    exprs: List["BooleanExpression"]


@dataclass
class AnyBooleanExpression:
    """Expression evaluating whether any given sub-expression hold."""

    exprs: List["BooleanExpression"]


@dataclass
class AssignedBooleanExpression:
    """Expression evaluating whether a given attribute is assigned for a given entity."""

    attribute_type: str
    reference: str


@dataclass
class AttributeBooleanExpression:
    """Expression evaluating an attribute assignment for a given boolean attribute and entity."""

    attribute_type: str
    reference: str


@dataclass
class AttributeDateExpression:
    """Expression evaluating an attribute assignment for a given date attribute and entity."""

    attribute_type: str
    reference: str


@dataclass
class AttributeDateTimeExpression:
    """Expression evaluating an attribute assignment for a given date-time attribute and entity."""

    attribute_type: str
    reference: str


@dataclass
class AttributeNumberExpression:
    """Expression evaluating an attribute assignment for a given number attribute and entity."""

    attribute_type: str
    reference: str


@dataclass
class AttributeStringExpression:
    """Expression evaluating an attribute assignment for a given string attribute and entity."""

    attribute_type: str
    reference: str


@dataclass
class AttributeTimeExpression:
    """Expression evaluating an attribute assignment for a given time attribute and entity."""

    attribute_type: str
    reference: str


class CmpOperator(Enum):
    """Operator for binary comparisons."""

    EQ = auto()
    GT = auto()
    GTE = auto()
    LT = auto()
    LTE = auto()
    NEQ = auto()


@dataclass
class DateCmpBooleanExpression:
    """Expression evaluating to a comparison between results of two date expressions."""

    lhs: "DateExpression"
    operator: CmpOperator
    rhs: "DateExpression"


@dataclass
class DateTimeCmpBooleanExpression:
    """Expression evaluating to a comparison between results of two date-time expressions."""

    lhs: "DateTimeExpression"
    operator: CmpOperator
    rhs: "DateTimeExpression"


@dataclass
class DirectlyLinkedToBooleanExpression:
    """Expression evaluating whether a given entity is directly linked to some other entity matching a condition."""

    alias: str
    condition: "BooleanExpression"
    entity_type: str
    link_alias: str
    source_id: int


@dataclass
class DirectLinkAggregateNumberExpression:
    """Expression evaluating a given aggregate across a given entity's direct links."""

    alias: str
    condition: "BooleanExpression"
    entity_type: str
    expr: "NumberExpression"
    link_alias: str
    op: AggregateOperator
    source_id: int


@dataclass
class IdInBooleanExpression:
    """Expression evaluating whether an entity's identifier occurs in a given list."""

    ids: List[str]
    reference: str


@dataclass
class IdStringExpression:
    """Expression evaluating to a given entity's identifier."""

    reference: str


@dataclass
class LinkedToBooleanExpression:
    """Expression evaluating whether a given entity is linked to some other entity matching a given condition."""

    alias: str
    condition: "BooleanExpression"
    entity_type: str
    source_id: int


@dataclass
class LinkAggregateNumberExpression:
    """Expression evaluating a given aggregate across a given entity's links."""

    alias: str
    condition: "BooleanExpression"
    entity_type: str
    expr: "NumberExpression"
    op: AggregateOperator
    source_id: int


@dataclass
class LinkAssignedBooleanExpression:
    """Expression evaluating whether a given relationship attribute is assigned for a given link."""

    link_attribute_type: str
    link_reference: str


@dataclass
class LinkAttributeBooleanExpression:
    """Expression evaluating an attribute assignment for a given boolean attribute and link."""

    link_attribute_type: str
    link_reference: str


@dataclass
class LinkAttributeDateExpression:
    """Expression evaluating an attribute assignment for a given date attribute and link."""

    link_attribute_type: str
    link_reference: str


@dataclass
class LinkAttributeDateTimeExpression:
    """Expression evaluating an attribute assignment for a given date-time attribute and link."""

    link_attribute_type: str
    link_reference: str


@dataclass
class LinkAttributeNumberExpression:
    """Expression evaluating an attribute assignment for a given number attribute and link."""

    link_attribute_type: str
    link_reference: str


@dataclass
class LinkAttributeStringExpression:
    """Expression evaluating an attribute assignment for a given string attribute and link."""

    link_attribute_type: str
    link_reference: str


@dataclass
class LinkAttributeTimeExpression:
    """Expression evaluating an attribute assignment for a given time attribute and link."""

    link_attribute_type: str
    link_reference: str


@dataclass
class LiteralBooleanExpression:
    """Expression evaluating to a given boolean literal."""

    boolean: bool


@dataclass
class LiteralDateExpression:
    """Expression evaluating to a given date literal."""

    date: date


@dataclass
class LiteralDateTimeExpression:
    """Expression evaluating to a given date-time literal."""

    date_time: datetime


@dataclass
class LiteralNumberExpression:
    """Expression evaluating to a given number literal."""

    number: float


@dataclass
class LiteralStringExpression:
    """Expression evaluating to a given string literal."""

    string: str


@dataclass
class LiteralTimeExpression:
    """Expression evaluating to a given time literal."""

    time: time


@dataclass
class MatchBooleanExpression:
    """Expression evaluating to a match between results of two given string expressions."""

    lhs: "StringExpression"
    mode: "MatchMode"
    operator: "MatchOperator"
    rhs: "StringExpression"


class MatchMode(Enum):
    """Case-sensitivity mode for matching two strings."""

    CASE_INSENSITIVE = auto()
    CASE_SENSITIVE = auto()


class MatchOperator(Enum):
    """Operator for matching two strings."""

    CONTAINS = auto()
    ENDS_WITH = auto()
    EQUALS = auto()
    STARTS_WITH = auto()


@dataclass
class NameStringExpression:
    """Expression evaluating to a given entity's name."""

    reference: str


@dataclass
class NotBooleanExpression:
    """Expression evaluating to the boolean negation of a given expression's result."""

    expr: "BooleanExpression"


@dataclass
class NumberCmpBooleanExpression:
    """Expression evaluating to a comparison between results of two number expressions."""

    lhs: "NumberExpression"
    operator: CmpOperator
    rhs: "NumberExpression"


NumberExpression = Union[
    AttributeNumberExpression,
    DirectLinkAggregateNumberExpression,
    LinkAggregateNumberExpression,
    LinkAttributeNumberExpression,
    LiteralNumberExpression,
]


@dataclass
class RelativeDateExpression:
    """Expression representing a date offset from a current timestamp."""

    days: int
    future: bool
    months: int
    years: int


DateExpression = Union[
    AttributeDateExpression,
    LinkAttributeDateExpression,
    LiteralDateExpression,
    RelativeDateExpression,
]


@dataclass
class RelativeDateTimeExpression:
    """Expression representing a date-time offset from a current timestamp."""

    days: int
    hours: int
    future: bool
    minutes: int
    months: int
    seconds: int
    years: int


DateTimeExpression = Union[
    AttributeDateTimeExpression,
    LinkAttributeDateTimeExpression,
    LiteralDateTimeExpression,
    RelativeDateTimeExpression,
]

StringExpression = Union[
    AttributeStringExpression,
    IdStringExpression,
    LinkAttributeStringExpression,
    LiteralStringExpression,
    NameStringExpression,
]


@dataclass
class TimeCmpBooleanExpression:
    """Expression evaluating to a comparison between results of two time expressions."""

    lhs: "TimeExpression"
    operator: CmpOperator
    rhs: "TimeExpression"


BooleanExpression = Union[
    ActiveBooleanExpression,
    AllBooleanExpression,
    AnyBooleanExpression,
    AssignedBooleanExpression,
    AttributeBooleanExpression,
    DateCmpBooleanExpression,
    DateTimeCmpBooleanExpression,
    DirectlyLinkedToBooleanExpression,
    IdInBooleanExpression,
    LinkAssignedBooleanExpression,
    LinkAttributeBooleanExpression,
    LinkedToBooleanExpression,
    LiteralBooleanExpression,
    MatchBooleanExpression,
    NotBooleanExpression,
    NumberCmpBooleanExpression,
    TimeCmpBooleanExpression,
]

TimeExpression = Union[
    AttributeTimeExpression,
    LinkAttributeTimeExpression,
    LiteralTimeExpression,
]
